from torchmetrics import Metric
import torch
from typing import Optional
from nlu_inference.utils import get_ents
from typing import List


class ChunkF1(Metric):

    full_state_update: Optional[bool] = False
    
    def __init__(self):
        super().__init__()
        self.add_state('correct', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_pred', default=torch.tensor(0.0), dist_reduce_fx='sum')
        self.add_state('all_true', default=torch.tensor(0.0), dist_reduce_fx='sum')

    def update(self, pred: List[List[str]], true: List[List[str]]):
        true_entities = set()
        for p in pred:
            for ent in get_ents(p):
                true_entities.add(ent)
        pred_entities = set()
        for t in true:
            for ent in get_ents(t):
                pred_entities.add(ent)
        self.correct += len(true_entities & pred_entities)
        self.all_pred += len(pred_entities)
        self.all_true += len(true_entities)

    def compute(self):
        return 2 * self.correct / (self.all_pred + self.all_true)


from wasabi import msg
from ._config import Config
import os
from pathlib import Path



class Push(object):
    """推送模型到模型保存服务器
    """
    def __init__(self):
        self.exp_file_path: Path = Config.EXP_SCRIPT
        self.project_dir_name = Path('.').resolve().name

    def cloud(self, 
              model_dir: str = '/data/wangshilin/24_push/model_folder',
              password: str = 'pachira123', 
              host: str = '192.168.130.27', 
              user: str = 'wangmengdi'):
        """推送云侧模型到保存目录, 云侧模型保存目录为: /data/wangshilin/24_push/model_folder/项目目录名/cloud/model

        Args:
            model_dir (str, optional): 模型保存的目录. Defaults to '/data/wangshilin/24_push/model_folder'.
            password (str, optional): 密码. Defaults to 'pachira123'.
            host (str, optional): 主机地址. Defaults to '192.168.130.27'.
            user (str, optional): 用户名. Defaults to 'wangmengdi'.
        """
        push_dir = f'{user}@{host}:{model_dir}/{self.project_dir_name}'
        result = os.system(f'expect {self.exp_file_path} {Config.CLOUD_DIR} {push_dir} {password}')
        if result != 0:
            msg.fail('push failed')
        else:
            msg.good(f'pushed {Config.CLOUD_DIR} to {push_dir}')
        
    def local(self,
              model_dir: str = '/data/wangshilin/24_push/model_folder',
              password: str = 'pachira123', 
              host: str = '192.168.130.27',
              user: str = 'wangmengdi'):
        """推送端侧模型到保存目录, 端侧模型保存目录为: /data/wangshilin/24_push/model_folder/项目目录名/local/models

        Args:
            model_dir (str, optional): 模型保存目录. Defaults to '/data/wangshilin/24_push/model_folder'.
            password (str, optional): 密码. Defaults to 'pachira123'.
            host (str, optional): 主机地址. Defaults to '192.168.130.27'.
            user (str, optional): 用户名. Defaults to 'wangmengdi'.
        """
        push_dir = f'{user}@{host}:{model_dir}/{self.project_dir_name}'
        result = os.system(f'expect {self.exp_file_path} {Config.LOCAL_DIR} {push_dir} {password}')
        if result != 0:
            msg.fail('push failed')
        else:
            msg.good(f'pushed {Config.LOCAL_DIR} to {push_dir}')
        
    def all(self,
            model_dir: str = '/data/wangshilin/24_push/model_folder',
            password: str = 'pachira123',
            host: str = '192.168.130.27',
            user: str = 'wangmengdi'):
        """推送端侧和云侧模型到保存目录, 端侧模型保存目录为: /data/wangshilin/24_push/model_folder/项目目录名/local/models, 云侧模型保存目录为: /data/wangshilin/24_push/model_folder/项目目录名/cloud/model

        Args:
            model_dir (str, optional): 模型保存目录. Defaults to '/data/wangshilin/24_push/model_folder'.
            password (str, optional): 密码. Defaults to 'pachira123'.
            host (str, optional): 主机地址. Defaults to '192.168.130.27'.
            user (str, optional): 用户名. Defaults to 'wangmengdi'.
        """
        self.cloud(model_dir, password, host, user)
        self.local(model_dir, password, host, user)
        
        


from pathlib import Path  
from dataclasses import dataclass


@dataclass
class Config:  
    # 项目默认配置文件目录
    DEFAULT_CONFIG_DIR = Path(__file__).parent / 'configs'
    # 项目默认语料文件目录
    DEFAULT_CORPUS_DIR = Path(__file__).parent / 'corpus'
    # 推送exp脚本文件路径
    EXP_SCRIPT = Path(__file__).parent / 'push.exp'
    
    # [language]
    ZH = ['zho', 'cmn']
    EN = ['eng']

    # [domain]
    # 项目默认fasttext正样本标签
    POS_LABEL: str = '1'
    # 项目默认fasttext负样本标签
    NEG_LABEL: str = '0'
    # domain 额外正样本文件名称,用于修复正样本bug以及测试
    POS_EXTRA: Path = Path('corpus', 'pos_extra.txt')
    # domain 额外负样本文件名称,用于修复负样本bug以及测试
    NEG_EXTRA: Path = Path('corpus', 'neg_extra.txt')

    # [intention]
    LABEL_PREFIX: str = '__label__'


    # [PATH]
    CLOUD_DIR = Path('checkpoints', 'cloud')
    # 云侧模型文件目录
    CLOUD_MODEL_DIR = Path('checkpoints', 'cloud', 'model')
    # 云侧nlu配置文件
    CLOUD_CONFIG_PATH = Path('checkpoints', 'cloud', 'model', 'config.cfg')
    # 云侧领域模型文件目录
    CLOUD_DOMAIN_DIR = Path('checkpoints', 'cloud', 'model', 'domain')
    # 云侧领域模型文件路径
    CLOUD_DOMAIN_MODEL = Path('checkpoints', 'cloud', 'model', 'domain', 'model.bin')
    # 云侧意图模型文件目录
    CLOUD_INTENTION_DIR = Path('checkpoints', 'cloud', 'model', 'intention')
    # 云侧意图模型文件路径
    CLOUD_INTENTION_MODEL = Path('checkpoints', 'cloud', 'model', 'intention', 'model.bin')
    # 云侧ner模型文件目录
    CLOUD_NER_DIR = Path('checkpoints', 'cloud', 'model', 'ner')
    # 云侧ner模型文件路径
    CLOUD_NER_MODEL = Path('checkpoints', 'cloud', 'model', 'ner', 'model.onnx')
    # 云侧词表文件
    CLOUD_VOCAB_PATH = Path('checkpoints', 'cloud', 'model', 'ner', 'word2id.json')
    # 云侧ner标签文件
    CLOUD_LABEL_PATH = Path('checkpoints', 'cloud', 'model', 'ner', 'id2label.json')
    # 端侧文件目录
    LOCAL_DIR = Path('checkpoints', 'local')
    # 端侧模型文件目录
    LOCAL_MODEL_DIR = Path("checkpoints", 'local', 'models')
    # 端侧nlu配置文件
    LOCAL_CONFIG_PATH = Path('checkpoints', 'local', 'models', 'config.cfg')
    # 端侧领域模型文件目录
    LOCAL_DOMAIN_DIR = Path("checkpoints", 'local', 'models', 'domain')
    # 端侧领域模型文件路径
    LOCAL_DOMAIN_MODEL = Path("checkpoints", 'local', 'models', 'domain', 'model.bin')
    # 端侧意图模型文件目录
    LOCAL_INTENTION_DIR = Path("checkpoints", 'local', 'models', 'intention')
    # 端侧意图模型文件路径
    LOCAL_INTENTION_MODEL = Path("checkpoints", 'local', 'models', 'intention', 'model.bin')
    # 端侧embedding模型文件路径
    LOCAL_EMBEDDING_MODEL = Path('checkpoints', 'local', 'embedding.onnx')
    # 端侧ner模型文件目录
    LOCAL_NER_DIR = Path("checkpoints", 'local', 'models', 'ner')
    # 端侧ner模型文件路径
    LOCAL_NER_MODEL = Path("checkpoints", 'local', 'models', 'ner', 'model.onnx')
    # 端侧词表文件
    LOCAL_VOCAB_PATH = Path('checkpoints', 'local', 'models', 'ner', 'word2id.json')
    # 端侧ner标签文件
    LOCAL_LABEL_PATH = Path('checkpoints', 'local', 'models', 'ner', 'id2label.json')


    # [train]

    # [dataset]
    DOMAIN_DATASET = Path('dataset', 'domain.txt')
    INTENTION_DATASET = Path('dataset', 'intention.txt')
    NER_DATASET = Path('dataset', 'ner')
    NER_PROCESSED_DATASET = Path('dataset', 'ner_processed')
    
    # [embedding]
    EN_EMBEDDING: str = 'glove-en-300d-20k'
    ZH_EMBEDDING: str = 'common-300d'
    



from pathlib import Path
from wasabi import msg
from nlu_inference.inference import FasttextInference, NERInferenceLocal, NERInferenceCloud
from nlu_inference.tokenizer import Tokenizer
from nlu_inference.io import CLSResult, NERResult
from ._config import Config
from nlp_data import NLUDocStore, NLUDocList
from nlp_data.document.nlu import NLUExample, NLUDoc
from typing import Optional, Literal, Union, Dict
from docarray.utils.filter import filter_docs
from .utils import get_tokenizer, get_domain_from_dir_name, get_lang_from_dir_name


def test_domain(domain_inference: FasttextInference, tokenizer: Tokenizer) -> None:
    """测试领域模型
    """
    badcases = []
    with open(Config.POS_EXTRA, 'r') as f:
        pos_lines = f.readlines()
    with open(Config.NEG_EXTRA, 'r') as f:
        neg_lines = f.readlines()
    for line in pos_lines:
        tokens = tokenizer(line, padding=False)[0]
        result: CLSResult = domain_inference(tokens=tokens)
        label = result.label 
        score = result.score
        if label != Config.POS_LABEL:
            msg.fail(f"领域模型badcase: {line} True: {Config.POS_LABEL} Predict: {label} Score: {score}")
            badcases.append(line)
    for line in neg_lines:
        tokens = tokenizer(line, padding=False)[0]
        result = domain_inference(tokens=tokens)
        label = result.label
        score = result.score
        if label != Config.NEG_LABEL:
            msg.fail(f"领域模型badcase: {line} True: {Config.NEG_LABEL} Predict: {label} Score: {score}")
            badcases.append(line)
    if len(badcases) == 0:
        msg.good("领域模型全部测试用例通过")
    else:
        msg.warn(f"领域模型badcase数量: {len(badcases)}")
            
def test_intention(intention_inference: FasttextInference, 
                   docs: NLUDocList,
                   tokenizer: Tokenizer) -> None:
    """测试意图模型
    """
    badcases = []
    for doc in docs:
        tokens = tokenizer(doc.text, padding=False)[0]
        result = intention_inference(tokens=tokens)
        label, score = result.label, result.score
        if label != doc.intention.text:
            msg.fail(f"意图模型badcase: {doc.text} True: {doc.intention.text} Predict: {label} Score: {score}")
            badcases.append(doc.text)
    if len(badcases) == 0:
        msg.good("意图模型全部测试用例通过")
    else:
        msg.warn(f"意图模型badcase数量: {len(badcases)}")
            
def test_ner(ner_inference: NERInferenceCloud, docs: NLUDocList, tokenizer: Tokenizer) -> None:
    badcases = []
    for doc in docs:
        doc: NLUDoc
        tokens = tokenizer(doc.text)[0]
        ner_results: NERResult = ner_inference(tokens=tokens)
        predict_doc = NLUDoc(text=doc.text)
        if len(ner_results) > 0:
            for label, text in ner_results.items():
                predict_doc.set_slot(text=text, label=label)
        example: NLUExample = NLUExample(x=doc, y=predict_doc)
        if example.is_slot_badcase:
            msg.fail(f"ner模型badcase: {doc.text} True: {[(slot.text, slot.label) for slot in doc.slots]} Predict: {[(slot.text, slot.label) for slot in predict_doc.slots]}")
            badcases.append(doc.text)
    if len(badcases) == 0:
        msg.good("ner模型全部测试用例通过")
    else:
        msg.warn(f"ner模型badcase数量: {len(badcases)}")
            
def test_docs(domain: str,
              tokenizer: Tokenizer,
              lang: Literal['zh', 'en'] = 'zh',
              domain_inference: Optional[FasttextInference] = None, 
              intention_inference: Optional[FasttextInference] = None, 
              ner_inference: Optional[NERInferenceLocal] = None) -> None:
    if lang == 'en':
        NLUDocStore.bucket_name = 'nlu-en'
    if domain_inference:
        test_domain(domain_inference, tokenizer=tokenizer)
    if intention_inference or ner_inference:
        all_doc_names = NLUDocStore.list(show_table=False)
        domain_test_doc_name = domain + '/test'
        if domain_test_doc_name not in all_doc_names:
            msg.warn(f"领域{domain}测试集不存在,将跳过意图和实体模型测试")
        else:
            domain_test_docs = NLUDocStore.pull(domain_test_doc_name, show_progress=False)
            if len(domain_test_docs) == 0:
                msg.warn(f"领域{domain}测试集为空")
            if intention_inference:
                test_intention(intention_inference, domain_test_docs, tokenizer=tokenizer)
            if ner_inference:
                test_ner(ner_inference, domain_test_docs, tokenizer=tokenizer)
        

class Test:
    """测试云侧和端侧模型
    """
    def __init__(self,
                 local_domain_path: Path = Config.LOCAL_DOMAIN_MODEL,
                 local_intention_path: Path = Config.LOCAL_INTENTION_MODEL,
                 local_ner_path: Path = Config.LOCAL_NER_MODEL,
                 local_embedding_model_path: Path = Config.LOCAL_EMBEDDING_MODEL,
                 local_vocab_path: Path = Config.LOCAL_VOCAB_PATH,
                 local_label_path: Path = Config.LOCAL_LABEL_PATH,
                 cloud_domain_path: Path = Config.CLOUD_DOMAIN_MODEL,
                 cloud_intention_path: Path = Config.CLOUD_INTENTION_MODEL,
                 cloud_ner_path: Path = Config.CLOUD_NER_MODEL,
                 cloud_vocab_path: Path = Config.CLOUD_VOCAB_PATH,
                 cloud_label_path: Path = Config.CLOUD_LABEL_PATH,
                 ) -> None:
        super().__init__()
        
        self.local_domain_path = Path(local_domain_path)
        self.local_intention_path = Path(local_intention_path)
        self.local_ner_path = Path(local_ner_path)
        self.local_embedding_model_path = Path(local_embedding_model_path)
        self.local_vocab_path = Path(local_vocab_path)
        self.local_label_path = Path(local_label_path)
        self.cloud_domain_path = Path(cloud_domain_path)
        self.cloud_intention_path = Path(cloud_intention_path)
        self.cloud_ner_path = Path(cloud_ner_path)
        self.cloud_vocab_path = Path(cloud_vocab_path)
        self.cloud_label_path = Path(cloud_label_path)
        self.domain = get_domain_from_dir_name()
        self.lang = get_lang_from_dir_name()
        
        
    def local(self, docs:bool = False, max_length: int = 32, pad_side: Literal['left', 'right'] = 'left'):
        """测试本地模型

        Args:
            max_length (int, optional): 最大文本长度,用于ner模型推理. Defaults to 32.
            pad_side (Literal[&#39;left&#39;, &#39;right&#39;], optional): 填充方式. Defaults to 'left'.
        """
        
        tokenizer = get_tokenizer(lang=self.lang, max_length=max_length, pad_side=pad_side)

        if not self.local_domain_path.exists():
            msg.warn(title=f"本地领域模型文件不存在: {self.local_domain_path}")
            domain_inference = None
        else:
            with msg.loading("加载本地领域模型..."):
                domain_inference = FasttextInference(self.local_domain_path)
        if not self.local_intention_path.exists():
            msg.warn(title=f"本地意图模型文件不存在: {self.local_intention_path}")
            intention_inference = None
        else:
            with msg.loading("加载本地意图模型..."):
                intention_inference = FasttextInference(self.local_intention_path)
        if not self.local_ner_path.exists() or not self.local_embedding_model_path.exists() or not self.local_vocab_path.exists() or not self.local_label_path.exists():
            msg.warn(title=f"本地ner模型文件不完整")
            ner_inference = None
        else:
            with msg.loading("加载本地ner模型..."):
                ner_inference = NERInferenceLocal(embedding_path=self.local_embedding_model_path,
                                                  model_path=self.local_ner_path,
                                                  label_path=self.local_label_path)
            
        if domain_inference is not None and intention_inference is not None and ner_inference is not None:
            msg.good(title="本地模型文件完整")
        
        if docs:
            test_docs(domain=self.domain, 
                    domain_inference=domain_inference, 
                    intention_inference=intention_inference, 
                    ner_inference=ner_inference,
                    tokenizer=tokenizer)
        
        while True:
            query = input("请输入测试query (输入q退出): ")
            tokens = tokenizer(query)[0]
            if query == 'q':
                break
            if domain_inference is not None:
                result = domain_inference(tokens=tokens)
                domain_label, score = result.label, result.score
                msg.info(f"Domain: {domain_label}, Score: {round(score, 2)}")
            if intention_inference is not None:
                result = intention_inference(tokens=tokens)
                intention_label, score = result.label, result.score
                msg.info(f"Intention: {intention_label}, Score: {round(score, 2)}")
            if ner_inference is not None:
                ner_result = ner_inference(tokens=tokens)
                msg.info(f"Enities: {ner_result}")
    
    def cloud(self, docs: bool = False, max_length: int = 32, pad_side: Literal['left', 'right'] = 'left'):
        """测试云侧模型

        Args:
            domain (str): 领域名称,用于加载测试docs
            max_length (int, optional): 最大文本长度. Defaults to 32.
            pad_side (Literal[&#39;left&#39;, &#39;right&#39;], optional): 填充方式. Defaults to 'left'.
        """
        
        tokenizer = get_tokenizer(lang=self.lang, max_length=max_length, pad_side=pad_side)

        if not self.cloud_domain_path.exists():
            msg.warn(title=f"云侧领域模型文件不存在: {self.cloud_domain_path}")
            domain_inference = None
        else:
            with msg.loading("加载云侧领域模型..."):
                domain_inference = FasttextInference(self.cloud_domain_path)
        if not self.cloud_intention_path.exists():
            msg.warn(title=f"云侧意图模型文件不存在: {self.cloud_intention_path}")
            intention_inference = None
        else:
            with msg.loading("加载云侧意图模型..."):
                intention_inference = FasttextInference(self.cloud_intention_path)
        if not self.cloud_ner_path.exists() or not self.cloud_vocab_path.exists() or not self.cloud_label_path.exists():
            msg.warn(title=f"云侧ner模型文件不完整")
            ner_inference = None
        else:
            with msg.loading("加载云侧ner模型..."):
                ner_inference = NERInferenceCloud(model_path=self.cloud_ner_path,label_path=self.cloud_label_path)
        if domain_inference is not None and intention_inference is not None and ner_inference is not None:
            msg.good(title="云侧模型文件完整")
        
        if docs:
            test_docs(domain=self.domain, 
                    domain_inference=domain_inference, 
                    intention_inference=intention_inference, 
                    ner_inference=ner_inference,
                    tokenizer=tokenizer)
            
        while True:
            query = input("请输入测试query (输入q退出): ")
            tokens = tokenizer(query)[0]
            if query == 'q':
                break
            if domain_inference is not None:
                domain_result = domain_inference(tokens=tokens)
                domain_label, score = domain_result.label, domain_result.score
                msg.info(f"Domain: {domain_label}, Score: {round(score, 2)}")
            if intention_inference is not None:
                intention_result = intention_inference(tokens=tokens)
                intention_label, score = intention_result.label, intention_result.score
                msg.info(f"Intention: {intention_label}, Score: {round(score, 2)}")
            if ner_inference is not None:
                ner_result = ner_inference(tokens=tokens)
                msg.info(f"Enities: {ner_result}")
                
    def add(self) -> None:
        """添加测试用例

        Args:
            domain (str): 领域名称,将根据其保存测试文件到nlp-data nlu bucket,例如 weather/test
        """
        domain = self.domain
        all_doc_names = NLUDocStore.list(show_table=False)
        train_doc_name = domain + '/train'
        if train_doc_name not in all_doc_names:
            msg.fail(f"领域不存在: {domain}. 请通过运行`nlp-data list --bucket nlu`查看所有领域", exits=True)
        with msg.loading(f"正在解析所有意图和实体标签"):
            train_docs = NLUDocStore.pull(train_doc_name, show_progress=False)
            all_intention_labels = sorted(set([i.text for i in train_docs.intention]))
            all_intention_labels = {i:label for i,label in enumerate(all_intention_labels)}
            all_ent_labels = sorted(set([ent.label for doc in train_docs for ent in doc.slots ]))
            all_ent_labels = {i:label for i,label in enumerate(all_ent_labels)}
        test_doc_name = domain + '/test'
        if test_doc_name not in all_doc_names:
            msg.info("当前测试用例为空")
            test_docs = NLUDocList()
            pre_num_test_docs = 0
        else:
            test_docs = NLUDocStore.pull(name=test_doc_name, show_progress=False)
            pre_num_test_docs = len(test_docs)
            msg.info(f"当前领域{domain}测试用例数量: {len(test_docs)}")
        while True:
            query = input("请输入测试query(输入q退出): ")
            if query == 'q':
                break
            if query in test_docs.text:
                msg.warn(f"`{query}` 该测试用例已存在。继续添加将替换原测试用例")
            new_doc = NLUDoc(text=query)
            intention_label_id = input(f"请选择意图标签(输入数字选择,输入q重新输入.): {all_intention_labels}")
            if intention_label_id == 'q':
                continue
            intention_label_id = int(intention_label_id)
            intention_label = all_intention_labels[intention_label_id]
            new_doc.set_intention(text=intention_label)
            msg.good(f"已选择意图标签: {intention_label}")
            while True:
                ent = input(f"请输入实体文本(输入q退出): ")
                if ent == 'q':
                    break
                if ent not in query:
                    msg.warn(f"实体`{ent}`不存在,请重新输入.")
                    continue
                ent_label_id = input(f"请选择实体标签(输入数字选择): {all_ent_labels}")
                ent_label_id = int(ent_label_id)
                ent_label = all_ent_labels[ent_label_id]
                new_doc.set_slot(text=ent, label=ent_label)
                msg.good(f"已为实体`{ent}`设置{ent_label}标签")
            ensure = input("请确认添加测试用例: 输入`1`确认, 输入其他字符放弃")
            if ensure == '1':
                # 删除已存在的测试用例
                query = {"text": {"$neq": query}}
                result_docs = filter_docs(docs=test_docs, query=query)
                if len(result_docs) < len(test_docs):
                    msg.info(f"已删除原有测试用例")
                result_docs.append(new_doc)
                test_docs = result_docs
                final_num_test_docs = len(test_docs)
                
        with msg.loading(f"正在更新测试用例: {test_doc_name}"):
                _ = NLUDocStore.push(docs=test_docs, name=test_doc_name, show_progress=False)
        msg.good(f"测试用例更新成功: {pre_num_test_docs} -> {final_num_test_docs}")
            
    def delete(self, text: Optional[str] = None) -> None:
        """删除测试用例

        Args:
            text (Optional[str], optional): 测试用例文本. Defaults to None.
        """
        domain = self.domain
        all_doc_names = NLUDocStore.list(show_table=False)
        test_doc_name = domain + '/test'
        if test_doc_name not in all_doc_names:
            msg.fail(f"领域{domain}测试用例不存在", exits=True)
        all_docs = NLUDocStore.pull(name=test_doc_name, show_progress=False)
        if text: 
            query = {"text": {"$neq": text}}
            result_docs = filter_docs(docs=all_docs, query=query)
            if len(result_docs) == len(all_docs):
                msg.warn(f"测试用例`{text}`不存在")
                return
            else:
                NLUDocStore.push(docs=result_docs, name=test_doc_name, show_progress=False)
                msg.good(f"测试用例`{text}`已删除")
        else:
            text = input("请输入要删除的测试用例: ")
            query = {"text": {"$neq": text}}
            result_docs = filter_docs(docs=all_docs, query=query)
            if len(result_docs) == len(all_docs):
                msg.warn(f"测试用例`{text}`不存在")
                return
            else:
                NLUDocStore.push(docs=result_docs, name=test_doc_name, show_progress=False)
                msg.good(f"测试用例`{text}`已删除")
        
    def list(self) -> None:
        """列出测试用例
        """
        domain = self.domain
        all_doc_names = NLUDocStore.list(show_table=False)
        test_doc_name = domain + '/test'
        if test_doc_name not in all_doc_names:
            msg.fail(f"领域{domain}测试用例不存在", exits=True)
        all_docs = NLUDocStore.pull(name=test_doc_name, show_progress=False)
        msg.info(f"领域{domain}测试用例数量: {len(all_docs)}")
        msg.info('-'*50)
        for doc in all_docs:
            msg.info(f"Query: {doc.text}")
            msg.info(f"Intention: {doc.intention.text}")
            msg.info(f"Entities: {[(ent.text, ent.label) for ent in doc.slots]}")
            msg.info('-'*50)
        


import lightning as L
from lightning.pytorch.utilities.types import EVAL_DATALOADERS, STEP_OUTPUT, TRAIN_DATALOADERS, OptimizerLRScheduler
import torch.nn as nn
from torch.utils.data import DataLoader
from nlp_data import EmbeddingDocStore
import numpy as np
import torch
from torch.optim import AdamW
from .layers import CRF, MultiDropout
from wasabi import msg
from datasets import load_from_disk, DatasetDict
from functools import lru_cache
from typing import Any, List, Literal, Optional
import pandas as pd
from nlu_inference.utils import ner_post_process
from .metrics import ChunkF1
from .utils import get_tokenizer
from .scheduler import linear_schedule_with_warmup
import json
import onnx
import os
from pathlib import Path
from typing import Union


class EntityExtractionDataModule(L.LightningDataModule):
    """实体抽取数据模块
    数据集格式: {"text":"这是一个长颈鹿","ents":[{"indices":[4,5,6],"label":"动物", "text":"长颈鹿"}]}
    """
        
    def __init__(self,
                 dataset_path: str,
                 processed_dataset_path: str,
                 embedding_name: str,
                 batch_size: int = 32,
                 max_length: int = 32,
                 pad_side: Literal['left', 'right'] = "left",
                 num_workers: int = 4,
                 lang: Literal['zho', 'eng', 'cmn'] = 'cmn',
                 process_batch_size: int = 10000,
                 **kwargs):
        super().__init__()
        self.save_hyperparameters(ignore='tokenizer')
        if Path(processed_dataset_path).exists():
            self.dataset = load_from_disk(processed_dataset_path)
            self.processed = True
        else:
            self.dataset = load_from_disk(self.hparams.dataset_path)
            self.processed = False
        self.hparams.id2bio = self.id2bio
        self.tokenizer = get_tokenizer(lang=lang, max_length=self.hparams.max_length, pad_side=self.hparams.pad_side)
        
    
    def prepare(self) -> None:
        if self.processed:
            pass
        else:
            processed_ds = self.dataset.map(self.bio_transform, batched=True, num_proc=self.hparams.num_workers, batch_size=self.hparams.process_batch_size)
            processed_dataset_dir = Path(self.hparams.processed_dataset_path)
            processed_ds.save_to_disk(processed_dataset_dir)
    
    
    def setup(self, stage: str = 'fit') -> None:
        if stage == 'fit':
            if self.processed:
                pass
            else:
                self.dataset = load_from_disk(self.hparams.processed_dataset_path)
         
    @property
    @lru_cache()
    def ent_labels(self) -> List:
        labels = sorted(pd.Series(np.concatenate(self.train_df['ents'])).apply(lambda x: x['label']).drop_duplicates().values)
        return labels
    
    @property
    def ent2id(self):
        return {l:i for i,l in enumerate(self.ent_labels)}

    @property
    def id2ent(self):
        return {i:l for l,i in self.ent2id.items()}
    
    @property
    @lru_cache()
    def bio_labels(self) -> List:
        b_labels = ['B' + '-' + l for l in self.ent_labels]
        i_labels = ['I' + '-' + l for l in self.ent_labels]
        return ['O'] + b_labels + i_labels
    
    @property
    def bio2id(self):
        return {l:i for i,l in enumerate(self.bio_labels)}
    
    @property
    def id2bio(self):
        return {i:l for l,i in self.bio2id.items()}
    
    @property
    @lru_cache()
    def train_df(self):
        return self.dataset['train'].to_pandas()
    
    @property
    @lru_cache()
    def val_df(self):
        return self.dataset['validation'].to_pandas()
    
    @property
    @lru_cache()
    def test_df(self):
        return self.dataset['test'].to_pandas()
    
    def get_batch_max_length(self, batch_text):
        return max([len(text) for text in batch_text])
    
    def bio_transform(self, examples):
        batch_text = examples['text']  
        inputs = {}
        batch_tokens = self.tokenizer(batch_text=batch_text)
        input_ids = [tokens.ids for tokens in batch_tokens]
        input_ids = torch.tensor(input_ids)
        inputs['input_ids'] = input_ids
        batch_tags = torch.zeros_like(input_ids, dtype=torch.long)
        for i, text in enumerate(batch_text):
            ents = examples['ents'][i]
            tokens = batch_tokens[i]
            for ent in ents:
                indices = ent['indices']
                start_token = tokens.char_to_token(indices[0])
                end_token = tokens.char_to_token(indices[-1])
                if start_token:
                    if not end_token:
                        end_token = len(tokens) - 1
                    batch_tags[i][start_token] = self.bio2id['B' + '-' + ent['label']]
                    if len(indices) > 1:
                        batch_tags[i][start_token+1: end_token+1] = self.bio2id['I' + '-' + ent['label']]
        inputs['tag_ids'] = batch_tags
        return inputs
    
    def train_dataloader(self) -> TRAIN_DATALOADERS:
        return DataLoader(dataset=self.dataset['train'], 
                          batch_size=self.hparams.batch_size,
                          num_workers=self.hparams.num_workers,
                          pin_memory=True,
                          collate_fn=self.collate_fn)
    
    def val_dataloader(self) -> EVAL_DATALOADERS:
        return DataLoader(dataset=self.dataset['validation'], 
                          batch_size=self.hparams.batch_size,
                          num_workers=self.hparams.num_workers,
                          pin_memory=True,
                          collate_fn=self.collate_fn)
        
    def collate_fn(self, batch):
        input_ids = torch.tensor([item['input_ids'] for item in batch], dtype=torch.long)
        tag_ids = torch.tensor([item['tag_ids'] for item in batch])
        return {'input_ids': input_ids, 'tag_ids': tag_ids}

class CRFClassifier(nn.Module):
    def __init__(self, 
                 n_tags: int, 
                 n_in: int, 
                 n_hidden: int):
        super().__init__()
        self.lstm = nn.LSTM(n_in, n_hidden, bidirectional=True, batch_first=True)
        self.dropout = MultiDropout()
        self.fc = nn.Linear(n_hidden * 2, n_tags)
        self.crf = CRF(num_tags=n_tags, batch_first=True)
        
    def forward(self, x):
        x, _ = self.lstm(x)
        x = self.dropout(x)
        x = self.fc(x)
        x = self.crf.decode(x)
        # 将输出的标签转化为one hot形式以跟之前模型兼容
        x = torch.nn.functional.one_hot(x, num_classes=self.crf.num_tags).float()
        
        return x
    
    def compute_loss(self, x, tags):
        x, _ = self.lstm(x)
        x = self.dropout(x)
        x = self.fc(x)
        loss = self.crf(emissions=x, tags=tags)
        return loss * -1
        

class CRFForEntityExtraction(L.LightningModule):
    def __init__(self, 
                 freeze_embedding: bool = False,
                 lr: float = 0.001,
                 hidden_size: int = 128,
                 **kwargs):
        super().__init__()
        self.save_hyperparameters()     
        
        self.num_labels = len(self.hparams.id2bio)
        self.embedding_docs = EmbeddingDocStore.pull(self.hparams.embedding_name)
        self.embeddings = nn.Embedding.from_pretrained(embeddings=torch.from_numpy(np.array([doc.embedding for doc in self.embedding_docs], dtype=np.float32)), freeze=self.hparams.freeze_embedding)
        self.classifier = CRFClassifier(n_tags=self.num_labels, n_in=self.embeddings.embedding_dim, n_hidden=self.hparams.hidden_size)
        self.hparams.tokenizer = get_tokenizer(lang=self.hparams.lang, max_length=self.hparams.max_length, pad_side=self.hparams.pad_side)
        
    def setup(self, stage: str):
        self.metric = ChunkF1()  
        
    def forward(self, input_ids):
        x = self.embeddings(input_ids)
        pred_ids = self.classifier(x)
        return pred_ids
    
    def compute_loss(self, input_ids, tags):
        x = self.embeddings(input_ids)
        loss = self.classifier.compute_loss(x=x, tags=tags)
        return loss
        
    def training_step(self, inputs) -> STEP_OUTPUT:
        loss = self.compute_loss(input_ids=inputs['input_ids'], tags=inputs['tag_ids'])
        self.log("train/loss", value=loss, prog_bar=True, on_step=True, on_epoch=False)
        return loss
    
    def validation_step(self, inputs) -> STEP_OUTPUT:
        pred_ids = self(input_ids=inputs['input_ids']).argmax(dim=-1)
        pred_labels = []
        for ids in pred_ids:
            pred_labels.append([self.hparams.id2bio[id] for id in ids.tolist()])
        true_label_ids = inputs['tag_ids']
        true_labels = []
        for i in range(len(true_label_ids)):
            indices = torch.where(true_label_ids[i]>=0)
            ids = true_label_ids[i][indices].tolist()
            true_labels.append([self.hparams.id2bio[id] for id in ids])
        self.metric(pred_labels, true_labels)
        self.log("val/f1", self.metric, on_step=False, on_epoch=True, prog_bar=True)
        
        
    def configure_optimizers(self) -> OptimizerLRScheduler:
        optimizer = AdamW(self.parameters(), lr=self.hparams.lr)
        num_training_steps = self.trainer.estimated_stepping_batches
        num_warmup_steps = num_training_steps // self.trainer.max_epochs // 5
        scheduler = linear_schedule_with_warmup(optimizer=optimizer, 
                                                num_training_steps=num_training_steps,
                                                num_warmup_steps=num_warmup_steps)
        return {"optimizer": optimizer, "lr_scheduler": {"interval": "step", "scheduler": scheduler}}
    
    
    def predict(self, text: str, device: str = 'cpu'):
        self.eval()
        self.freeze()
        self.to(device)
        tokens = self.hparams.tokenizer(text)[0]
        input_ids = torch.tensor(tokens.ids).unsqueeze(0).to(device)
        outputs = self(input_ids=input_ids.to(self.device))
        bio_tags = [self.hparams.id2bio[id] for id in outputs[0].tolist()]
        return ner_post_process(labels=bio_tags, tokens=tokens)
    
    
    def save_label_dict(self, save_path: str):
        with open(save_path, 'w') as f:
            json.dump(self.hparams.id2bio, f)
    
    
    def to_onnx(self, 
                save_path: str,
                embedding_save_path: Optional[str] = None,
                max_length: int = 32,
                input_name: str = 'input', 
                output_name: str = 'output', 
                opset_version: int = 14,
                split_crf: bool = True) -> None:
        self.eval()
        self.freeze()
        text = "好" * max_length
        batch_tokens = self.hparams.tokenizer(text)
        input_ids = [tokens.ids for tokens in batch_tokens]
        input_ids = torch.tensor(input_ids)
        if split_crf:
            with torch.no_grad():
                logits = self.embeddings(input_ids)
                torch.onnx.export(self.classifier, 
                                logits, 
                                save_path, 
                                opset_version=opset_version,
                                input_names=[input_name], 
                                output_names=[output_name],
                                dynamic_axes={input_name: {0: 'batch_sizel', 1: 'seq_length'}, output_name: {0: 'batch_size', 1: 'seq_length'}})
                # 更改形状的最后一维为标签大小, 不然端侧模型会报错
                ner_model = onnx.load(str(save_path)) 
                ner_model.graph.input[0].type.tensor_type.shape.dim[1].dim_value  = max_length
                ner_model.graph.output[0].type.tensor_type.shape.dim[2].dim_value  = len(self.hparams.id2bio)
                onnx.save(ner_model, save_path)
                
                torch.onnx.export(
                    self.embeddings,
                    input_ids,
                    embedding_save_path,
                    opset_version=opset_version,
                    input_names=[input_name],
                    output_names=[output_name],
                    dynamic_axes={input_name: {0: 'batch_size', 1: 'seq_len'}}
                )
        else:
            with torch.no_grad():
                torch.onnx.export(self, 
                                input_ids, 
                                save_path, 
                                opset_version=opset_version,
                                input_names=[input_name], 
                                output_names=[output_name],
                                # dynamic_axes={input_name: {0: 'batch_size', 1: 'seq_len'}, output_name: {0: 'batch_size', 1: 'seq_len'}})
                                dynamic_axes={input_name: {0: 'batch_size'}, output_name: {0: 'batch_size', 1: 'seq_length'}})


from torch.optim.lr_scheduler import LambdaLR


def linear_schedule_with_warmup(optimizer, 
                                num_warmup_steps, 
                                num_training_steps, 
                                last_epoch=-1):
    """Linear warmup and then linear decay.
    """
    
    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps)))

    return LambdaLR(optimizer=optimizer, lr_lambda=lr_lambda, last_epoch=last_epoch)


from pathlib import Path
from fasttext import train_supervised
from wasabi import msg
from .model import CRFForEntityExtraction, EntityExtractionDataModule
from typing import Any, List, Literal, Union, Optional
import lightning as L
from lightning.pytorch.callbacks import ModelCheckpoint, EarlyStopping, RichProgressBar
from lightning.pytorch.loggers import CSVLogger
from datetime import datetime
from ._config import Config
from .utils import print_config, get_default_embedding_name, get_lang_from_dir_name, get_domain_from_dir_name, get_default_nlu_config
import warnings
import onnxruntime as ort
import torch
import torch._dynamo
torch._dynamo.config.suppress_errors = True



class Train():
    """训练云侧和端侧模型
    """
    def __init__(self,
                 local_save_dir: Path = Config.LOCAL_DIR,
                 local_domain_dir: Path = Config.LOCAL_DOMAIN_DIR,
                 local_intention_dir: Path = Config.LOCAL_INTENTION_DIR,
                 local_embedding_path: Path = Config.LOCAL_EMBEDDING_MODEL,
                 local_ner_dir: Path = Config.LOCAL_NER_DIR,
                 cloud_save_dir: Path = Config.CLOUD_DIR,
                 cloud_model_dir: Path = Config.CLOUD_MODEL_DIR,
                 cloud_intention_dir: Path = Config.CLOUD_INTENTION_DIR,
                 cloud_domain_dir: Path = Config.CLOUD_DOMAIN_DIR,
                 cloud_ner_dir: Path = Config.CLOUD_NER_DIR,
                 domain_dataset: Path = Config.DOMAIN_DATASET,
                 intention_dataset: Path = Config.INTENTION_DATASET,
                 ner_dataset: Path = Config.NER_DATASET,
                 processed_ner_dataset: Path = Config.NER_PROCESSED_DATASET,
                 cloud_config_path: Path = Config.CLOUD_CONFIG_PATH,
                 local_config_path: Path = Config.LOCAL_CONFIG_PATH
                 ) -> None:
        super().__init__()
        # Path
        self.local_save_dir = local_save_dir
        self.local_domain_dir = local_domain_dir
        self.local_intention_dir = local_intention_dir
        self.local_ner_dir = local_ner_dir
        self.local_embedding_path = local_embedding_path
        self.cloud_save_dir = cloud_save_dir
        self.cloud_model_dir = cloud_model_dir
        self.cloud_intention_dir = cloud_intention_dir
        self.cloud_domain_dir = cloud_domain_dir
        self.cloud_ner_dir = cloud_ner_dir
        self.cloud_config_path = cloud_config_path
        self.local_config_path = local_config_path
        
        # Dataset
        self.domain_dataset = str(domain_dataset)
        self.intention_dataset = str(intention_dataset)
        self.ner_dataset = ner_dataset
        self.processed_ner_dataset = processed_ner_dataset
        
        self.lang = get_lang_from_dir_name()
        self.domain_name = get_domain_from_dir_name()
        

    def domain(self, 
               epochs: int = 30, 
               lr: float = 0.1, 
               ngrams: int = 2,
               bucket: int = 50000,
               dim: int = 100,
               quantize: bool = True,
               qnorm: bool = False,
               retrain: bool = True,
               retrain_epoch: int = 1,
               cutoff: int = 50000,
               qout: bool = False,
               save_local: bool = True,
               save_cloud: bool = True,
               **kwargs):
        """训练领域二分类模型
        """
        print_config(locals())
        model = train_supervised(self.domain_dataset, 
                                 epoch=epochs, 
                                 lr=lr, 
                                 dim=dim,
                                 wordNgrams=ngrams, 
                                 bucket=bucket,
                                 **kwargs)    
        if quantize:
            with msg.loading(f"正在评估原始模型..."):
                result = model.test(self.domain_dataset)
            msg.info(f"原始模型评估结果 -> 测试数量: {result[0]} 准确率: {round(result[1], 4)} 召回率: {round(result[2], 4)}")
            
            with msg.loading(f"正在量化领域模型"):
                model.quantize(input=self.domain_dataset, 
                               qnorm=qnorm, 
                               retrain=retrain, 
                               epoch=retrain_epoch, 
                               cutoff=cutoff, 
                               qout=qout)   
                
            with msg.loading(f"正在评估量化模型..."):
                quantize_result = model.test(self.domain_dataset)
            msg.info(f"量化模型评估结果 -> 测试数量: {quantize_result[0]} 准确率: {round(quantize_result[1], 4)} 召回率: {round(quantize_result[2], 4)}")
            
            if save_cloud:
                if not self.cloud_domain_dir.exists():
                    self.cloud_domain_dir.mkdir(parents=True)
                model.save_model(str(self.cloud_domain_dir / 'model.bin'))
                msg.good(f"云端领域量化模型保存成功: {self.cloud_domain_dir / 'model.bin'}")
            if save_local:
                if not self.local_domain_dir.exists():
                    self.local_domain_dir.mkdir(parents=True)
                model.save_model(str(self.local_domain_dir / 'model.bin'))
                msg.good(f"本地领域量化模型保存成功: {self.local_domain_dir / 'model.bin'}")
        else:    
            if save_cloud:
                if not self.cloud_domain_dir.exists():
                    self.cloud_domain_dir.mkdir(parents=True)
                model.save_model(str(self.cloud_domain_dir / 'model.bin'))
                msg.good(f"云端领域模型保存成功: {self.cloud_domain_dir / 'model.bin'}")
            if save_local:
                if not self.local_domain_dir.exists():
                    self.local_domain_dir.mkdir(parents=True)
                model.save_model(str(self.local_domain_dir / 'model.bin'))
                msg.good(f"本地领域模型保存成功: {self.local_domain_dir / 'model.bin'}")
    
    
    def intention(self,
                  epochs: int = 30,
                  lr: float = 0.1,
                  ngrams: int = 2,
                  bucket: int = 100000,
                  dim: int = 100,
                  quantize: bool = True,
                  qnorm: bool = False,
                  retrain: bool = True,
                  retrain_epoch: int = 1,
                  cutoff: int = 100000,
                  qout: bool = False,
                  save_cloud: bool = True,
                  save_local: bool = True,
                  **kwargs):
        """训练意图分类模型
        """
        print_config(locals())
        model = train_supervised(self.intention_dataset,
                                 epoch=epochs,
                                 lr=lr,
                                 dim=dim,
                                 wordNgrams=ngrams,
                                 bucket=bucket,
                                 **kwargs)
        if quantize:
            with msg.loading(f"正在评估原始模型..."):
                result = model.test(self.intention_dataset)
            msg.info(f"原始模型评估结果 -> 测试数量: {result[0]} 准确率: {round(result[1], 4)} 召回率: {round(result[2], 4)}")
            
            with msg.loading(f"正在量化意图模型"):
                model.quantize(input=self.intention_dataset,
                               qnorm=qnorm,
                               retrain=retrain,
                               epoch=retrain_epoch,
                               cutoff=cutoff,
                               qout=qout)
            with msg.loading(f"正在评估量化模型..."):
                quantize_result = model.test(self.intention_dataset)
            msg.info(f"量化模型评估结果 -> 测试数量: {quantize_result[0]} 准确率: {round(quantize_result[1], 4)} 召回率: {round(quantize_result[2], 4)}")

            if save_cloud:
                if not self.cloud_intention_dir.exists():
                    self.cloud_intention_dir.mkdir(parents=True)
                model.save_model(str(self.cloud_intention_dir / 'model.bin'))
                msg.good(f"云端意图量化模型保存成功: {self.cloud_intention_dir / 'model.bin'}")
            if save_local:
                if not self.local_intention_dir.exists():
                    self.local_intention_dir.mkdir(parents=True)
                model.save_model(str(self.local_intention_dir / 'model.bin'))
                msg.good(f"本地意图量化模型保存成功: {self.local_intention_dir / 'model.bin'}")
        else:
            if save_cloud:
                if not self.cloud_intention_dir.exists():
                    self.cloud_intention_dir.mkdir(parents=True)
                model.save_model(str(self.cloud_intention_dir / 'model.bin'))
                msg.good(f"云端意图模型保存成功: {self.cloud_intention_dir / 'model.bin'}")
            if save_local:
                if not self.local_intention_dir.exists():
                    self.local_intention_dir.mkdir(parents=True)
                model.save_model(str(self.local_intention_dir / 'model.bin'))
                msg.good(f"本地意图模型保存成功: {self.local_intention_dir / 'model.bin'}")
        
        
    def ner(self, 
            lr: float = 0.001,
            devices: Union[List, int] = [0],
            hidden_size: int = 128,
            max_epochs: int = 10,
            freeze_embedding: bool = True,
            embedding_name: Optional[str] = None,
            batch_size: int = 32,
            max_length: int = 32,
            pad_side: Literal['left', 'right'] = "left",
            num_workers: int = 4,
            patience: int = 3,
            save_cloud: bool = True,
            save_local: bool = True,
            label_save_name: str = 'id2label.json',
            vocab_save_name: str = 'word2id.json',
            ignore_warnings: bool = True,
            optimize: bool = False,
            process_batch_size: int = 5000,
            compile: bool = False):
        """训练ner模型

        Args:
            lr (float, optional): 学习率. Defaults to 0.001.
            devices (Union[List, int], optional): 训练的gpu设备. Defaults to [0].
            hidden_size (int, optional): lstm的隐层温度. Defaults to 128.
            max_epochs (int, optional): 最大训练迭代次数. Defaults to 10.
            freeze_embedding (bool, optional): 是否冻结embedding层. Defaults to True.
            embedding_name (Optional[str], optional): 从nlpdata下载的embedding名称,默认会自动选择. Defaults to None.
            batch_size (int, optional): 训练一个批次的大小. Defaults to 32.
            max_length (int, optional): 一个训练样本的最大token数量. Defaults to 32.
            pad_side (Literal[&#39;left&#39;, &#39;right&#39;], optional): 批次样本填充方式. Defaults to "left".
            num_workers (int, optional): dataloader数据处理进程数. Defaults to 4.
            patience (int, optional): 提早结束训练的条件,当训练的f1分数连续patience次不增加时会停止训练. Defaults to 3.
            save_cloud (bool, optional): 是否保存云侧模型. Defaults to True.
            save_local (bool, optional): 是否保存端侧模型. Defaults to True.
            label_save_name (str, optional): 保存的ner标签文件. Defaults to 'id2label.json'.
            vocab_save_name (str, optional): ner模型embedding对应的词表. Defaults to 'word2id.json'.
            ignore_warnings (bool, optional): 是否忽略提醒. Defaults to True.
            optimize (bool, optional): 是否优化onnx模型. Defaults to False.
            process_batch_size (int, optional): 数据处理的一个批次大小. Defaults to 5000.
            compile (bool, optional): 是否编译模型,成功开启编译的时候会增加训练速度. Defaults to False.
        """
        print_config(locals())
        
        L.seed_everything(42)
        
        if ignore_warnings:
            warnings.filterwarnings("ignore")
        
        early_stoppin = EarlyStopping(monitor='val/f1',
                                      patience=patience, 
                                      mode='max')
        # 日期时间为目录名
        log_dir: Path = Path('logs', 'ner', datetime.now().strftime("%Y-%m-%d-%H-%M"))
        ckpt_dir = log_dir / 'checkpoints'

        ckpt = ModelCheckpoint(dirpath=ckpt_dir, 
                            monitor="val/f1",
                            filename="epoch{epoch}-f1_{val/f1:.3f}", 
                            auto_insert_metric_name=False,
                            mode='max',
                            save_last=False,
                            save_top_k=1)
        rich_bar = RichProgressBar()

        logger = CSVLogger(save_dir=log_dir, name='csv', version='')
        
        trainer = L.Trainer(devices=devices, 
                            max_epochs=max_epochs, 
                            precision='16-mixed',
                            accelerator='gpu', 
                            callbacks=[early_stoppin, ckpt, rich_bar],
                            logger=logger,
                            profiler="simple")
        lang = self.lang
        if not embedding_name:
            embedding_name = get_default_embedding_name(lang)
            msg.info(f"{lang}默认embedding: {embedding_name}")
        
        dm = EntityExtractionDataModule(dataset_path=self.ner_dataset, 
                                        processed_dataset_path=self.processed_ner_dataset,
                                        embedding_name=embedding_name,
                                        batch_size=batch_size,
                                        process_batch_size=process_batch_size,
                                        max_length=max_length,
                                        pad_side=pad_side,
                                        num_workers=num_workers,
                                        lang=lang)
        if trainer.local_rank == 0:
            dm.prepare()
            
        
        model = CRFForEntityExtraction(freeze_embedding=freeze_embedding,
                                       lr=lr,
                                       hidden_size=hidden_size,
                                       **dm.hparams)
        
        if compile:
            model = torch.compile(model=model)
            
        trainer.fit(model=model, datamodule=dm)
        
        if save_cloud:
            if not self.cloud_ner_dir.exists():
                self.cloud_ner_dir.mkdir(parents=True)
            crf_model_path = self.cloud_ner_dir / 'model.onnx'
            model.to_onnx(save_path=crf_model_path, split_crf=False, max_length=max_length)
            model.save_label_dict(self.cloud_ner_dir / label_save_name)
            model.hparams.tokenizer.save_vocab_json(self.cloud_ner_dir / vocab_save_name)
            msg.good(f"云侧ner模型保存成功: {crf_model_path}")
            
            if optimize:
                with msg.loading(f"正在优化onnx模型..."):
                    sess_options = ort.SessionOptions()
                    # Set graph optimization level
                    sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
                    # To enable model serialization after graph optimization set this
                    sess_options.optimized_model_filepath = str(crf_model_path)
                    session = ort.InferenceSession(crf_model_path, sess_options)
                
            nlu_config = get_default_nlu_config(max_length=max_length, pad_side=pad_side, is_cloud=True)
            nlu_config.to_disk(self.cloud_config_path)
            msg.good(f"云侧nlu配置文件保存成功: {self.cloud_config_path}")
            
        if save_local:
            if not self.local_ner_dir.exists():
                self.local_ner_dir.mkdir(parents=True)
            crf_model_path = self.local_ner_dir / 'model.onnx'
            model.to_onnx(save_path=crf_model_path, split_crf=True, max_length=max_length, embedding_save_path=self.local_embedding_path)
            model.save_label_dict(self.local_ner_dir / label_save_name)
            model.hparams.tokenizer.save_vocab_json(self.local_ner_dir / vocab_save_name)
            msg.good(f"端侧ner模型保存成功: {crf_model_path}")
            
            if optimize:
                with msg.loading(f"正在优化onnx模型..."):
                    sess_options = ort.SessionOptions()
                    # Set graph optimization level
                    sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
                    # To enable model serialization after graph optimization set this
                    sess_options.optimized_model_filepath = str(crf_model_path)
                    session = ort.InferenceSession(crf_model_path, sess_options)
                
            nlu_config = get_default_nlu_config(max_length=max_length, pad_side=pad_side, is_cloud=False)
            nlu_config.to_disk(self.local_config_path)
            msg.good(f"端侧nlu配置文件保存成功: {self.local_config_path}")


from jsonargparse import CLI 
from .prepare import Prepare
from .train import Train
from .test import Test
from .push import Push
from pathlib import Path
import shutil
from wasabi import msg
from ._config import Config
from typing import Literal
from .utils import check_dir_name

AVALIABLE_LANG = Literal['zho', 'eng', 'cmn']


def init_project(domain: str, lang: AVALIABLE_LANG):
    """在本目录下初始化项目,并生成语料文件,默认模型训练配置文件,默认项目配置文件以及数据文件。

    Args:
        name (str): 项目名称,会在本目录下创建一个同名的文件夹
    """
    project_dir = Path(domain + '_' + lang)
    check_dir_name(project_dir)
    if project_dir.exists():
        msg.fail(f'{project_dir.stem} already exists', exits=1)
    project_dir.mkdir()
    msg.good(f'created project dir {project_dir.absolute()}')
    shutil.copytree(Config.DEFAULT_CORPUS_DIR, project_dir / 'corpus')
    shutil.copytree(Config.DEFAULT_CONFIG_DIR, project_dir / 'configs')
    # 创建模型文件路径
    cloud_model_dir = project_dir / Config.CLOUD_MODEL_DIR
    local_model_dir = project_dir / Config.LOCAL_MODEL_DIR
    if not cloud_model_dir.exists():
        cloud_model_dir.mkdir(parents=True)
    if not local_model_dir.exists():
        local_model_dir.mkdir(parents=True)
    

        
components = {
    "init": init_project,
    "prepare": Prepare,
    "train": Train,
    "test": Test,
    "push": Push
}

def run_cli():
    CLI(components=components, as_positional=False)


from typing import List, Tuple, Optional, Literal, Union, Dict
from nlp_data import NLUDocStore, EmbeddingDocStore, EmbeddingDocList
from nlu_inference.tokenizer import CharTokenizer, WordTokenizer, Tokenizer
import rich
import rich.tree
import rich.syntax
from pathlib import Path
from ._config import Config
from nlu_inference.config import Config as NLUConfig


def check_domain_exist_in_bucket(domain: str) -> bool:
    """检查领域是否存在

    Args:
        domain (str): 领域名称.

    Returns:
        bool: 是否存在.
    """
    doc_save_name = domain + '/train'
    all_doc_names = NLUDocStore.list(show_table=False)
    if doc_save_name in all_doc_names:
        return True
    else:
        return False
    
    

def print_config(config: dict):
    """打印配置文件

    Args:
        config (dict): 配置文件.
    """
    style = 'dim'
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)
    if 'self' in config:
        config.pop('self')
    for k, v in config.items():
        if k == 'kwargs':
            for k1, v1 in v.items():
                branch = tree.add(k1)
                branch.add(rich.syntax.Syntax(str(v1), "yaml"))
        else:
            branch = tree.add(k)
            branch.add(rich.syntax.Syntax(str(v), "yaml"))
    rich.print(tree)
    


def get_tokenizer(lang: Literal['zho', 'eng', 'cmn'], 
                  max_length: int = 32, 
                  pad_side: Literal['left', 'right'] = 'left',
                  **kwargs) -> Tokenizer:
    """根据语种获取默认分词器

    Args:
        lang (Literal[&#39;zh&#39;, &#39;en&#39;]): 语种
        vocab_path (Union[str, Path, Dict[str, int]]): 词表路径或者词表字典.
        max_length (int, optional): 最大token长度. Defaults to 32.
        pad_side (Literal[&#39;left&#39;, &#39;right&#39;], optional): 填充方式. Defaults to 'left'.
    """
    embedding_name = get_default_embedding_name(lang=lang)
    embedings: EmbeddingDocList = EmbeddingDocStore.pull(embedding_name, show_progress=False)
    vocab = embedings.get_vocab()
    if lang in Config.ZH:
        return CharTokenizer(vocab=vocab, max_length=max_length, pad_side=pad_side)
    if lang in Config.EN:
        return WordTokenizer(vocab=vocab, max_length=max_length, pad_side=pad_side)
    else:
        raise ValueError(f'unknow language {lang}')
    
    
def get_default_embedding_name(lang: Literal['zho', 'eng', 'cmn']) -> str:
    """获取默认embedding名称

    Args:
        lang (str): 语种.

    Returns:
        str: embedding名称.
    """
    if lang in Config.ZH:
        return Config.ZH_EMBEDDING
    if lang in Config.EN:
        return Config.EN_EMBEDDING
    else:
        raise ValueError(f'unknow language {lang}')
    
    
def get_lang_from_dir_name() -> str:
    """从目录名获取语种

    Args:
        dir_name (str): 目录名.

    Returns:
        str: 语种.
    """
    dir_name = Path.cwd().stem
    lang = dir_name.split('_')[-1]
    return lang
    
    
def get_domain_from_dir_name() -> str:
    """从目录名获取领域

    Args:
        dir_name (str): 目录名.

    Returns:
        str: 领域.
    """
    dir_name = Path.cwd().stem
    splits = dir_name.split('_')
    assert len(splits) == 2, f'文件夹名字不符合规范,请使用领域_语种的格式,比如weather_zho, time-query_eng'
    return splits[0]


def check_dir_name(dir: Path):
    """检查目录名是否符合规范

    Args:
        dir (Path): 目录.
    """
    dir_name = dir.stem
    splits = dir_name.split('_')
    if len(splits) != 2:
        raise ValueError(f'文件夹名字不符合规范,请使用领域_语种的格式,比如weather_zho, time-query_eng') 
    lang = splits[-1]
    if lang not in Config.ZH and lang not in Config.EN:
        raise ValueError(f"unknow language {lang}, available language is {Config.ZH} and {Config.EN}") 
    
def get_default_tokenizer_type(lang: str = Literal['zho', 'cmn', 'eng']):
    """获取默认分词器类型

    Returns:
        str: 分词器类型.
    """
    if lang in Config.ZH:
        return 'char'
    if lang in Config.EN:
        return 'word'
    else:
        raise ValueError(f'unknow language {lang}')
    
    
def get_default_nlu_config(max_length: int = 32,
                           pad_side: Literal['left', 'right'] = 'left',
                           pad_id: int = 0,
                           unk_token: str = "UNK",
                           is_cloud: bool = True):
    """获取默认nlu配置
    """
    lang = get_lang_from_dir_name()
    domain = get_domain_from_dir_name()
    tokenizer_type = get_default_tokenizer_type(lang)
    config_str = """
    [nlu]
    @languages = cmn
    domain = "schedule"

    [nlu.domain_inference]
    @inferences = "domain.fasttext"
    model_path = ${path.domain}

    [nlu.intention_inference]
    @inferences = "intention.fasttext"
    model_path = ${path.intention}

    [nlu.ner_inference]
    @inferences = "ner.onnx.cloud"
    model_path = ${path.ner}
    label_path = ${path.label}

    [nlu.tokenizer]
    @tokenizers = "char"
    max_length = 32
    vocab = ${path.vocab}
    pad_side = "left"
    pad_id = 0
    unk_token = "UNK"

    [path]
    ner = "model/ner/model.onnx"
    vocab = "model/ner/word2id.json"
    label = "model/ner/id2label.json"
    domain = "model/domain/model.bin"
    intention = "model/intention/model.bin"
    """
    overrides = {"nlu.domain": domain, 
                 "nlu.@languages": lang, 
                 "nlu.tokenizer.@tokenizers": tokenizer_type, 
                 "nlu.tokenizer.max_length": max_length, 
                 "nlu.tokenizer.pad_side": pad_side, 
                 "nlu.tokenizer.pad_id": pad_id, 
                 "nlu.tokenizer.unk_token": unk_token}
    if not is_cloud:
        overrides['path.ner'] = "models/ner/model.onnx"
        overrides['path.embedding'] = "embedding.onnx"
        overrides['path.vocab'] = "models/ner/word2id.json"
        overrides['path.label'] = "models/ner/id2label.json"
        overrides['path.domain'] = "models/domain/model.bin"
        overrides['path.intention'] = "models/intention/model.bin"
        overrides['nlu.ner_inference.@inferences'] = 'ner.onnx.local'
        overrides['nlu.ner_inference.embedding_path'] = '${path.embedding}'
        overrides['nlu.ner_inference.model_path'] = '${path.ner}'
        overrides['nlu.ner_inference.label_path'] = '${path.label}'
        
    config = NLUConfig().from_str(config_str, overrides=overrides, interpolate=False)
    return config
        
    


from pathlib import Path
from wasabi import msg
from nlp_data import NLUDocList, NLUDocStore
from typing import Optional
from random import choices
import os
import pandas as pd
from .utils import print_config, get_lang_from_dir_name, get_domain_from_dir_name, get_tokenizer
from ._config import Config
import shutil


class Prepare():
    def __init__(self, 
                 corpus_dir: str = 'corpus', 
                 dataset_dir: str = 'dataset',
                 pos_file: str = 'positive.txt',
                 neg_file: str = 'negative.txt',
                 pos_extra: str = 'pos_extra.txt', 
                 neg_extra: str = 'neg_extra.txt',
                 pos_label: str = '__label__1#',
                 neg_label: str = '__label__0#',
                 domain_dataset: str = 'domain.txt',
                 intention_dataset: str = 'intention.txt',
                 ner_dataset: str = 'ner',
                 ner_processed_dataset: str = 'ner_processed'):
        """数据准备

        Args:
            corpus_dir: 语料目录,包含abnf,以及正负语料文件等.
            dataset_dir: 数据集目录,包含ner,intention,domain数据集.
            test_dir: 测试目录,包含domain,intention,ner测试数据文件.
            pos_file: 领域正例文件名称.
            neg_file: 领域反例文件名称.
            pos_extra: 领域正例额外文件名称.
            neg_extra: 领域反例额外文件名称.
            pos_label: 领域正例标签.
            neg_label: 领域反例标签.
            domain_dataset: 领域数据集文件名称.
            intention_dataset: 意图数据集文件名称.
            ner_dataset: ner数据集文件名称.
        """
        super().__init__()
        self.corpus_dir = Path(corpus_dir)
        if not self.corpus_dir.exists():
            self.corpus_dir.mkdir(parents=True)
        self.pos_extra = Path(self.corpus_dir, pos_extra)
        if not self.pos_extra.exists():
            self.pos_extra.touch()
        self.neg_extra = Path(self.corpus_dir, neg_extra)
        if not self.neg_extra.exists():
            self.neg_extra.touch()
        self.pos_file = Path(self.corpus_dir, pos_file)
        if not self.pos_file.exists():
            self.pos_file.touch()
        self.neg_file = Path(self.corpus_dir, neg_file)
        if not self.neg_file.exists():
            self.neg_file.touch()
        self.pos_label = pos_label
        self.neg_label = neg_label
        self.abnf_out_dir = Path(corpus_dir) / 'abnf' / 'out'
        if not self.abnf_out_dir.exists():
            self.abnf_out_dir.mkdir(parents=True)
        self.abnf_dir = Path(corpus_dir) / 'abnf'
        self.dataset_dir = Path(dataset_dir)
        self.domain_dataset = Path(dataset_dir, domain_dataset)
        self.intention_dataset = Path(dataset_dir, intention_dataset)
        self.ner_dataset = Path(dataset_dir, ner_dataset)
        self.ner_processed_dataset = Path(dataset_dir, ner_processed_dataset)
        self.store = NLUDocStore
        self.lang = get_lang_from_dir_name()
        if self.lang in Config.EN:
            self.store.bucket_name = 'nlu-en' 
        self.domain = get_domain_from_dir_name()
        # 获取默认tokenizer,用于生成数据集之前的预处理,预处理时候不会填充
        self.tokenizer = get_tokenizer(lang=self.lang, max_length=100, pad_side='right')
        
        
    def docs(self, num_docs: int = 5000, parse: bool = True, push: bool = True):
        """准备文档数据,并推送到nlp-data

        Args:
            num_docs (int): 生成的文档数量。
            parse (bool, optional): 是否分析文档. Defaults to True.
            push (bool, optional): 是否推送. Defaults to True.
        """
        print_config(locals())
        if self.abnf_out_dir.exists():
            for file in self.abnf_out_dir.iterdir():
                file.unlink()
            self.abnf_out_dir.rmdir()
        os.system(f"cd {str(self.abnf_dir)} && bash run.sh . {num_docs}")
        add_space = True if (self.lang in Config.EN) else False
        docs = NLUDocList.from_abnf_output(abnf_output_dir=self.abnf_out_dir, domain=self.domain, add_space=add_space)
        docs.set_language(self.lang)
        if parse:
            # 分析意图标签数量
            intention_labels = set([intention.text for intention in docs.intention])
            intention_labels = {label: 0 for label in intention_labels}
            # 统计标签数量
            for intention in docs.intention:
                intention_labels[intention.text] += 1
            msg.info(f"所有意图标签数量: {len(intention_labels)}")
            all_intention_labels = [item for item in intention_labels.items()]
            intention_df = pd.DataFrame.from_records(all_intention_labels,columns=['标签', '数量'])
            # 分析意图标签数量
            intention_mean = int(intention_df['数量'].mean())
            msg.info(f"当前意图标签数量均值: {intention_mean}")
            intention_dis = abs(intention_df['数量'] - intention_mean).mean()
            intention_top = int(intention_mean + intention_dis)
            intention_bottom = int(intention_mean - intention_dis)
            # 暂时不对数据量太多的意图进行提示
            # upper_df = intention_df[intention_df['数量']>intention_top]
            # for i, row in upper_df.iterrows():
            #     msg.warn(f"意图{row['标签']}: 当前{row['数量']}条, 数量过多")
            lower_df = intention_df[intention_df['数量']<intention_bottom]
            for i, row in lower_df.iterrows():
                msg.warn(f"意图{row['标签']}: 当前{row['数量']}条, 数量过少")
            if len(lower_df) == 0:
                msg.good(f"意图标签数据分布正常")
            # 分析实体标签数量
            all_slot_labels = [slot.label for slots in docs.slots if slots is not None for slot in slots]
            slot_labels = set(all_slot_labels)
            slot_labels = {label: 0 for label in slot_labels}
            for slot in all_slot_labels:
                slot_labels[slot] += 1
            msg.info(f"所有实体标签数量: {len(slot_labels)}")
            all_slot_labels = [item for item in slot_labels.items()]
            slot_df = pd.DataFrame.from_records(all_slot_labels,columns=['标签', '数量'])
            slot_mean = int(slot_df['数量'].mean())
            msg.info(f"当前实体标签数量均值: {slot_mean}")
            slot_dis = abs(slot_df['数量'] - slot_mean).mean()
            slot_top = int(slot_mean + 3 * slot_dis)
            slot_bottom = int(slot_mean - 3 * slot_dis)
            # upper_df = slot_df[slot_df['数量']>slot_top]
            # for i, row in upper_df.iterrows():
            #     msg.warn(f"实体{row['标签']}: 当前{row['数量']}条, 数量过多")
            lower_df = slot_df[slot_df['数量']<slot_bottom]
            for i, row in lower_df.iterrows():
                msg.warn(f"实体{row['标签']}: 当前{row['数量']}条, 数量过少")
            if len(lower_df) == 0:
                msg.good(f"各类实体数据分布正常")
        if push:
            # 推送到nlp-data
            doc_save_name = self.domain + '/train'
            all_doc_names = self.store.list(show_table=False)
            if doc_save_name in all_doc_names:
                msg.warn("已存在同名文档,将覆盖")
            self.store.push(docs=docs, name=doc_save_name)
            msg.info(f"已推送到nlp-data: {doc_save_name}, 可通过命令行查看: nlp-data list --bucket {self.store.bucket_name}")
        
        
    def corpus(self, num_choices: Optional[int] = None):
        """准备domain正负向数据,将nlp-data领域数据作为正向,其他领域数据作为负向

        Args:
            num_docs (int): 生成文档数量.
            num_choices (Optional[int], optional): 负向每个领域的数量. 默认为None,此时将按照正向领域数量的2倍作为负向领域数量.
        """
        print_config(locals())
        msg.info(f"prepare corpus for domain: {self.domain}, from bucket {self.store.bucket_name}")
        # 首先删除原有文件
        if self.neg_file.exists():
            self.neg_file.unlink()
        if self.pos_file.exists():
            self.pos_file.unlink()
        domain = self.domain
        # 获取所有文档名称  
        all_doc_names = self.store.list(show_table=False)
        neg_doc_names = set([doc.split('/')[0] for doc in all_doc_names if domain not in doc])
        pos_doc_names = set([doc.split('/')[0] for doc in all_doc_names if domain in doc])
        num_domain = len(pos_doc_names) + len(neg_doc_names)
        
        # 保存正向数据
        pos_docs = NLUDocList()
        for pos_doc_name in pos_doc_names:
            train_pos_doc_name = pos_doc_name + '/train'
            if train_pos_doc_name in all_doc_names:
                docs = self.store.pull(name=train_pos_doc_name)
                pos_docs.extend(docs)
            test_pos_doc_name = pos_doc_name + '/test'
            if test_pos_doc_name in all_doc_names:
                docs = self.store.pull(name=test_pos_doc_name)
                pos_docs.extend(docs)
        with open(self.pos_file, 'a', encoding='utf-8') as f:
            for doc in pos_docs:
                f.write(doc.text+'\n')
        msg.good(f'已保存{len(pos_docs)}条正例')
        
        # 计算每个领域的反例数量
        if not num_choices:
            num_choices = (2 * len(pos_docs)) // (num_domain - len(pos_doc_names))
            msg.info(f'每个领域的反例数量为2 * {len(pos_docs)} / ({num_domain} - {len(pos_doc_names)}) = {num_choices}')
        else:
            msg.info(f'每个领域的反例数量为{num_choices}')
        # 保存负向数据
        for neg_doc_name in neg_doc_names:
            neg_docs = NLUDocList()
            train_neg_doc = neg_doc_name + '/train'
            if train_neg_doc in all_doc_names:
                docs = self.store.pull(train_neg_doc)
                neg_docs.extend(docs=docs)
            test_neg_doc = neg_doc_name + '/test'
            if test_neg_doc in all_doc_names:
                docs = self.store.pull(test_neg_doc)
                neg_docs.extend(docs=docs)
            # 如果意图中没有none,则按照意图进行采样
            if None not in neg_docs.intention:
                neg_docs = neg_docs.sample_by_intention(n_sample=num_choices)
            else:
                neg_docs = choices(neg_docs, k=num_choices)
            with open(self.neg_file, 'a', encoding='utf-8') as f:
                for doc in neg_docs:
                    f.write(doc.text)
            msg.good(f'已保存{neg_doc_name}的{num_choices}条反例')
            
    def dataset(self, up_sample: bool = False, domain: bool = True, intention: bool = True, ner: bool = True):
        """准备数据集,领域数据集会添加pos_extra和neg_extra的数据
        
        Args:
            up_sample (bool, optional): 是否对意图数据集进行上采样. Defaults to False.
            domain (bool, optional): 是否保存领域数据集. Defaults to True.
            intention (bool, optional): 是否保存意图数据集. Defaults to True.
            ner (bool, optional): 是否保存ner数据集. Defaults to True.
        """
        print_config(locals())
        
        domain_name = self.domain
        docs: NLUDocList = self.store.pull(domain_name + '/train')
        
        # 保存意图数据集
        if intention or ner:
            # 如果需要上采样,则进行上采样
            if up_sample:
                pre_num = len(docs)
                with msg.loading(f"正在根据意图数量上采样文档..."):
                    docs = docs.up_sampling_by_intention()
                msg.good(f"已上采样文档: {pre_num} -> {len(docs)}")   
                
            if ner:
                dsd = docs.convert_slots_to_ner_dataset()
                if self.ner_dataset.exists():
                    shutil.rmtree(self.ner_dataset)
                # 删除处理之前的数据集
                if self.ner_processed_dataset.exists():
                    shutil.rmtree(self.ner_processed_dataset)
                dsd.save_to_disk(self.ner_dataset)
                msg.good(f"NER数据集已保存至{self.ner_dataset}")
                
            if intention:
                if self.lang in Config.EN:
                    for doc in docs:
                        tokens = self.tokenizer(doc.text, padding=False)[0]
                        doc.text = ' '.join(tokens.text)
                    docs.convert_intention_to_fasttext_dataset(save_path=self.intention_dataset, split_type="no_split")
                else:
                    docs.convert_intention_to_fasttext_dataset(save_path=self.intention_dataset, split_type="char")
                
                msg.good(f"意图数据集已保存至{self.intention_dataset}")
                
        if domain:
            # 保存领域数据集
            if self.domain_dataset.exists():
                self.domain_dataset.unlink()
            with open(self.pos_file, 'r', encoding='utf-8') as f:
                pos_lines = f.readlines()
            with open(self.pos_extra, 'r', encoding='utf-8') as f:
                pos_extra_lines = f.readlines()
                pos_lines += pos_extra_lines
                msg.info(f"正例额外数据数量: {len(pos_extra_lines)}")
            with open(self.neg_file, 'r', encoding='utf-8') as f:
                neg_lines = f.readlines()
            with open(self.neg_extra, 'r', encoding='utf-8') as f:
                neg_extra_lines = f.readlines()
                neg_lines += neg_extra_lines
                msg.info(f"反例额外数据数量: {len(neg_extra_lines)}")
                
            with open(self.domain_dataset, 'a', encoding='utf-8') as f:
                for line in pos_lines:
                    line = line.strip()
                    if line:
                        if self.lang in Config.EN:
                            line = " ".join(self.tokenizer(line, padding=False)[0].text)
                        else:
                            line = " ".join(list(line))
                        f.write(f'{self.pos_label} {line}\n')
                for line in neg_lines:
                    line = line.strip()
                    if line:
                        if self.lang in Config.EN:
                            line = " ".join(self.tokenizer(line, padding=False)[0].text)
                        else:
                            line = " ".join(list(line))
                        f.write(f'{self.neg_label} {line}\n')
            msg.good(f"领域数据集已保存至{self.domain_dataset}")